/*
  Prolog code generator class.
  (c) Gertjan van Noord 1997.
  based on work by Matt Phillips 1996.

  maybe to do:

pereira@alta.research.att.com 
Fernando Pereira
11th November 1994 

Here's a technique that I have used in Quintus Prolog. It works,
although it is on the slow side. In the DCG, instead of using [X] for
a terminal symbol X, use `X, with the following operator and predicate
definitions.

:- op(500, fx, `).

`(C, @(P0,Stream), @(P,Stream)) :-
  stream_position(Stream, _, P0),
  get0(Stream, C),
  stream_position(Stream, P).

If the grammar start symbol is S and the file to parse is File, define:

parse_from_file(S, File) :-
  open(File, read, Stream),
  stream_position(Stream, Pos0),
  phrase(S, @(Pos0,Stream), @(Pos, _)),
  stream_position(Stream, _, Pos),
  get0(Stream, C), is_end_of_file(C),    % optional: file is completely parsed
  close(Stream).

The relative slowness is due to the constant fiddling with stream
positions, which involves going through several layers of Prolog, C,
I/O libraries and system calls. But using assert to implement some
form of streams would be even worse.

*/

#include "PrologCodeGenerator.h"


PrologCodeGenerator::PrologCodeGenerator (const FooObject &s,
					  istream &cf, 
					  ostream &b,
					  int option_c,
					  int option_u,
					  int option_l
					  ) :
  scandef (s), codefrags (cf, "prolog"), body (b), errors (0),
  option_compact_code (option_c), option_use_module (option_u),
  option_lookahead (option_l)

{
  const string &scannerName = scandef.getAttrString ("name");

  doStart (scannerName);

  doEdgeLists ();

  doStates ();

  doProdBodies (scannerName);
}

void PrologCodeGenerator::doStart (const string &scannerName)
{
  body << "%% tokenizer in Prolog generated by elex.
";

  if (option_use_module)
    {
      body << "
:- module('" << scannerName << "',
    [ tokenize/2, 
      tokenize/4,
      tokenize_input/1,
      tokenize_input/2,
      tokenize_input_per_line/1,
      tokenize_input_per_line/2,
      tokenize_input_per_line_with_findall/1,
      tokenize_input_per_line_with_findall/2
    ]).
";
    }
  body << "
%% tokenize(Chars,Tokens).
%% the string Chars (list of character codes)
%% is tokenized as the list of atoms Tokens.
tokenize([],[]).
tokenize([Char|Chars0],Tokens0) :-
    (   tokenize(Tokens0,Tokens1,[Char|Chars0],Chars1,Goal),
        look_ahead(Chars1," << option_lookahead << ")
    ->  call(Goal),          % stick to this symbol.
                             % the effect is longest prefix match, because the 
                             % final state cases below are ordered after the
                             % continuation cases ..
        tokenize(Chars1,Tokens1)
    ;   error(Char,Chars0,Tokens0)
    ).

tokenize(Tokens0,Tokens,Chars0,Chars) :-
    tokenize(Tokens0,Tokens,Chars0,Chars,Goal),
    call(Goal).

%% tokenize(Tokens0,Tokens,Chars0,Chars,Goal).
%% finds the next token in the list of characters Chars0. Remaining characters
%% is Chars, Goal is to be called.
tokenize(Tokens0,Tokens,String0,String,Goal):-
";
  if (option_compact_code)
    body << 
      "	tokenize(0,Tokens0,Tokens,String0,String,Symbol,[],Symbol,Goal).
";
  else	
    body << 
      "	tokenize0(Tokens0,Tokens,String0,String,Symbol,[],Symbol,Goal).
";
  body << "
%% return the list of tokens for the characters read from standard
%% input (up to end_of_file).
tokenize_input(Tokens) :-
    tokenize_input(user,Tokens).

%% return the list of tokens for the characters read from File.
%% note: the input is first read in into a list of character codes.
%% This list of char codes is then tokenized. This is neccesary
%% because tokenization is non-deterministic, and of course we can't
%% undo read (get0) commands upon backtracking.  
tokenize_input(File,Tokens) :-
    see(File),
    read_buffer(Chars),
    seen,
    tokenize(Chars,Tokens).

read_buffer(Chars) :-
    get0(Char),
    read_buffer(Char,Chars).

read_buffer(-1,Chars) :-
    !, Chars=[].
read_buffer(Char,[Char|Chars]) :-
    get0(Char1),
    read_buffer(Char1,Chars).

% read lines from stdin. TokenList is a list where each element is
% the list of tokens (one list of tokens per line).
tokenize_input_per_line(Tokens) :-
    tokenize_input_per_line(user,Tokens).

% read lines from File. TokenList is a list where each element is the 
% list of tokens (one list of tokens per line).
tokenize_input_per_line(File,TokensList) :-
    see(File),
    tokenize_input_per_line0(TokensList),
    seen.

tokenize_input_per_line0([Tokens|TokensList]) :-
    read_line(Chars,End),
    tokenize(Chars,Tokens),
    tokenize_input_per_line1(End,TokensList).

tokenize_input_per_line1(-1,T) :-
    !,
    T=[].
tokenize_input_per_line1(_,T) :-
    tokenize_input_per_line0(T).

% as tokenize_input_per_line, with different implementation.
% this version uses findall rather than recursive defintion.
% Is faster for SWI prolog.
tokenize_input_per_line_with_findall(Tokens) :-
    tokenize_input_per_line_with_findall(user,Tokens).

tokenize_input_per_line_with_findall(File,TokensList) :-
    see(File),
    findall(Tokens,read_and_tokenize_line(Tokens),TokensList),
    seen.

read_and_tokenize_line(Tokens) :-
    repeat,
    read_line(Chars,End),
    tokenize(Chars,Tokens),
    (   End =  -1
    ->  !
    ;   true
    ).

read_line(Line,End) :-
    get0(Char),
    read_line(Char,Line,End).

read_line(-1,R,E) :- !, R=[], E= -1.   % end of file
read_line(10,R,E) :- !, R=[], E= 10.   % end of line
read_line(C,[C|T],E) :-
    get0(C1),
    read_line(C1,T,E).

look_ahead([],_).
look_ahead([H|T],N) :- 
    look_aheadX(N,[H|T]).

look_aheadX(0,_) :- !.
look_aheadX(N0,[H|T]) :-
    tokenize(_,_,[H|T],Str,_),     % ignore goal during look-ahead
    N is N0-1,
    look_ahead(Str,N).


";

}

// do the edge list definitions for all states
void PrologCodeGenerator::doEdgeLists ()
{
  const FooObject::Objects &states = scandef.getAttrs ("states");

  for (FooObject::Objects::Iterator s (states); s; s++)
  {
    const string &stateNo = s.ref ().getAttrString ("state");
    const FooObject::Objects &edges = s.ref ().getAttrs ("edges");

    // output any edges for this state
    if (edges.nItems () > 0)
    {
      doEdgeList (edges, stateNo);      
    }
  }
}

// do <edges> for state <stateNo>
// for each transition from P to Q with Sym, add the clause
// tokenize(P,T0,T,[Sym|S0],S,[Sym|Sy0],Sy,Token):-
//    tokenize(Q,T0,T,S0,S,Sy0,Sy,Token).
// However, all transitions for a state are collected in a single
// disjunction.
void PrologCodeGenerator::doEdgeList (const FooObject::Objects &edges,
				   const string &stateNo)
{
  if (option_compact_code)
    body << "tokenize(" << stateNo << 
      ",Ts0,Ts,[C|S0],S,[C|Sy0],Sy,T,G):-\n    (";
  for (FooObject::Objects::Iterator e (edges); e; )
    {
      const FooObject &edge = e.ref ();
      const string &range = edge.getAttrString ("range");
      
      if (option_compact_code)
	{
	  if (range.length () > 1)
	    {
	      body << "   C >= " << (int)range [0];
	        if ( (int)range [2] != -1)
		  body << ", C =< " << (int)range [2];

		body << " ->\n    " <<
		"tokenize(" << edge.getAttrString ("target") <<
		",Ts0,Ts,S0,S,Sy0,Sy,T,G)\n    ;";
	    }
	  else
	    {
	      body << "   C =:= " << (int)range [0] << " ->\n    " <<
		"tokenize(" << edge.getAttrString ("target") << 
		",Ts0,Ts,S0,S,Sy0,Sy,T,G)\n    ;";
	    }
	}
      else
	{
	  int i = range [0];
	  int j = (range.length () > 1 ? range [2] : range [0]);


	  // Hm. This is a hack. What is the principled way of doing this?
	  // Why 255? 
	  if (j == -1)
	    j = 255;
	  
	  // note: separate clauses for each symbol in an interval...
	  for (;j >= i;++i)
	    {
	      body << "tokenize" << stateNo << "(";
	      body << (int)i;
	      body << ",Ts0,Ts,S0,S,Sy0,Sy,T,G):-\n";
	      body << "    tokenize" << edge.getAttrString ("target");
	      body << "(Ts0,Ts,S0,S,Sy0,Sy,T,G).\n";
	    }
	}
      e++;
    }
  if (option_compact_code)
    body << "   fail\n    ).\n\n";
}

// do the states table
// for each final state P with production name Name, add a clause
// tokenize(P,Ts0,Ts,S,S,Sy,Sy,Token) :- 'Name'(Token,Ts0,Ts).
void PrologCodeGenerator::doStates ()
{
  
  const FooObject::Objects &states = scandef.getAttrs ("states");

  for (FooObject::Objects::Iterator s (states); s; )
  {
    const FooObject &state = s.ref ();
    const string &stateNo = state.getAttrString ("state");
    const string &prodName = state.getAttrString ("production");

    if (!option_compact_code)
      {
	// state with outgoing transitions
	// this if-statement is ordered before the next one, to have
	// a preference for longer matches.
	if (state.hasAttr ("edges"))
	  {
	    body << "tokenize" << stateNo <<"(Ts0,Ts,[C|S0],S,[C|Sy0],Sy,T,G):-\n";
	    body <<	"    tokenize" << stateNo << "(C,Ts0,Ts,S0,S,Sy0,Sy,T,G).\n"; 
	  }
      }

    // final state
    if (prodName.length () > 0)
      {
	if (option_compact_code)
	  {
	    body << "tokenize(" << stateNo <<
	      ",__Tokens0,__Tokens,S,S,Sy,Sy,__Token,";
	    body << "'" << prodName << 
	      "'(__Token,__Tokens0,__Tokens)).\n\n";
	  }
	else
	  {
	    body << "tokenize" << stateNo <<"(__Tokens0,__Tokens,S,S,Sy,Sy,__Token,";
	    body << "'" << prodName << "'(__Token,__Tokens0,__Tokens)).\n\n";
	  }
      }
    s++;
  }
}


// inserts the production bodies from the code frags
// for each production name Name, add the clause
// 'Name'(__Token,__Tokens0,__Tokens) :- <code frag>
// NB: the variable names are crucial.
void PrologCodeGenerator::doProdBodies (const string &scannerName)
{
  const FooObject::Objects &productions =
    scandef.getAttrs ("productions");

  codefrags.nextFrag ();

  for (FooObject::Objects::Iterator p (productions); p; p++)
  {
    if ( p.ref ().getAttr () == "error" )
      body << "error(__Char,__Chars,__Tokens) :-\n";
    else
      body << "'" << p.ref ().getAttr () << "'(__Token,__Tokens0,__Tokens):-\n";

    codefrags.nextFrag ();

    if (codefrags.hasFrag ())
      codefrags.outputFrag (body);
    else
    {
      cerr << "warning: no code for production " <<
	p.ref ().getAttr () << endl;

      body << "__Tokens0=__Tokens. %% *** Missing code fragment\n";
    }
    body << endl;
  }
}
